@@ -70,6 +70,9 @@ static unsigned char *buf;
 static int imtu = 672;
 static int omtu = 0;
 
+/* Default FCS option */
+static int fcs = 0x01;
+
 /* Default data size */
 static long data_size = -1;
 static long buffer_size = 2048;
@@ -220,6 +223,8 @@ static int do_connect(char *svr)
 	if (rfcmode > 0)
 		opts.mode = rfcmode;
 
+	opts.fcs = fcs;
+
 	if (setsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, sizeof(opts)) < 0) {
 		syslog(LOG_ERR, "Can't set L2CAP options: %s (%d)",
 							strerror(errno), errno);
@@ -380,6 +385,8 @@ static void do_listen(void (*handler)(int sk))
 	if (rfcmode > 0)
 		opts.mode = rfcmode;
 
+	opts.fcs = fcs;
+
 	if (setsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, sizeof(opts)) < 0) {
 		syslog(LOG_ERR, "Can't set L2CAP options: %s (%d)",
 							strerror(errno), errno);
@@ -1030,12 +1037,13 @@ static void usage(void)
 		"\t[-b bytes] [-i device] [-P psm]\n"
 		"\t[-I imtu] [-O omtu]\n"
 		"\t[-L seconds] enable SO_LINGER\n"
-		"\t[-F seconds] enable deferred setup\n"
+		"\t[-W seconds] enable deferred setup\n"
 		"\t[-B filename] use data packets from file\n"
 		"\t[-N num] send num frames (default = infinite)\n"
 		"\t[-C num] send num frames before delay (default = 1)\n"
 		"\t[-D milliseconds] delay after sending num frames (default = 0)\n"
 		"\t[-X mode] select retransmission/flow-control mode\n"
+		"\t[-F fcs] use CRC16 check (default = 1)\n"
 		"\t[-R] reliable mode\n"
 		"\t[-G] use connectionless channel (datagram)\n"
 		"\t[-A] request authentication\n"
@@ -1052,7 +1060,7 @@ int main(int argc, char *argv[])
 
 	bacpy(&bdaddr, BDADDR_ANY);
 
-	while ((opt=getopt(argc,argv,"rdscuwmnxyzpb:i:P:I:O:B:N:L:F:C:D:X:RGAESMT")) != EOF) {
+	while ((opt=getopt(argc,argv,"rdscuwmnxyzpb:i:P:I:O:B:N:L:W:C:D:X:F:RGAESMT")) != EOF) {
 		switch(opt) {
 		case 'r':
 			mode = RECV;
@@ -1136,7 +1144,7 @@ int main(int argc, char *argv[])
 			linger = atoi(optarg);
 			break;
 
-		case 'F':
+		case 'W':
 			defer_setup = atoi(optarg);
 			break;
 
@@ -1163,6 +1171,10 @@ int main(int argc, char *argv[])
 				rfcmode = atoi(optarg);
 			break;
 
+		case 'F':
+			fcs = atoi(optarg);
+			break;
+
 		case 'R':
 			reliable = 1;
 			break;
